import re
from collections import deque
from typing import Deque
from cfn_check.shared.types import (
    Data,
    Items,
    YamlList,
    YamlObject,
    YamlValueBase,
)


numbers_pattern = re.compile(r'\d+')

def search(
    resources: YamlObject,
    path: str,
):
    items: Items = deque()
    items.append(resources)

    segments = path.split("::")[::-1]
    # Queries can be multi-segment,
    # so we effectively perform per-segment
    # repeated DFS searches, returning the matches
    # for each segment

    composite_keys: list[str] = []

    while len(segments):
        query = segments.pop()
        items, keys = search_with_query(items, query)

        if len(composite_keys) == 0:
            composite_keys.extend(keys)

        else:
            updated_keys: list[str] = []
            for composite_key in composite_keys:
                while len(keys):
                    key = keys.pop()

                    updated_keys.append(f'{composite_key}.{key}')

            composite_keys = updated_keys

    results: list[tuple[str, Data]] = []
    for idx, item in enumerate(list(items)):
        results.append((
            composite_keys[idx],
            item,
        ))

    return results


def parse_list_query(query: str):
    
    queries = query.strip('[]').split(',')

    if len(queries) < 1:
        return None
    
    indexes = []
    for query in queries:

        if match := numbers_pattern.match(query):
            indexes.append(
                int(match.group(0))
            )
    
    return indexes


def parse_list_matches(
    query: str,
    node: YamlList,
):
    if indexes := parse_list_query(query):
        return [
            item
            for idx, item in enumerate(node)
            if idx in indexes
        ]

    return [
        str(idx) for idx in indexes
    ], node


def search_with_query(
    items: Items,
    query: str,
) -> tuple[Items, Deque[str]]:
    
    found: Items = deque()

    keys: Deque[str] = deque()
    
    while len(items):
        node = items.pop()

        key: (
            str | None
        ) = None
        value: (
            YamlValueBase | YamlList | YamlObject | None
        ) = None

        if isinstance(node, dict):
            items.extend(node.items())

        elif query.startswith('[') and query.startswith(']'):
            indexes, matched = parse_list_matches(query, node)
            keys.extend(indexes)
            found.extend(matched)
 
        elif isinstance(node, list):
            items.extend(node)

        elif isinstance(node, tuple): 
            key, value = node

        else:
            # If we encounter just
            # a raw YAML int/bool/etc
            # then we should just
            # skip to the next iteration
            continue

        if (
            key == query or query == "*"
        ) and value:
            keys.append(key)
            found.append(value)
        
    return found, keys



from cfn_check import Rules, Rule


class ValidateResourceTypes(Rules):

    @Rule(
        "Resources::*::Type",
        "It checks Resource::Type is correctly definined",
    )
    def validate_test(self, value: str): 
        assert value is not None, '❌ Resource Type not defined'
        assert isinstance(value, str), '❌ Resource Type not a string'



import inspect

from async_logging import LogLevelName, Logger, LoggingConfig
from cocoa.cli import CLI, ImportType

from cfn_check.cli.utils.attributes import bind
from cfn_check.cli.utils.files import load_templates
from cfn_check.evaluation.validate import run_validations
from cfn_check.logging.models import InfoLog
from cfn_check.rules.rules import Rules
from cfn_check.validation.validator import Validator


@CLI.command()
async def validate(
    path: str,
    file_pattern: str | None = None,
    rules: ImportType[Rules] = None,
    tags: list[str] = [
        'Ref',
        'Sub',
        'Join',
        'Select',
        'Split',
        'GetAtt',
        'GetAZs',
        'ImportValue',
        'Equals',
        'If',
        'Not',
        'And',
        'Or',
        'Condition',
        'FindInMap',
    ],
    log_level: LogLevelName = 'info',
):
    '''
    Validate Cloud Foundation
    
    @param rules Path to a file containing Rules
    @param file_pattern A string pattern used to find template files
    @param tags List of CloudFormation intrinsic function tags
    @param log_level The log level to use
    '''

    logging_config = LoggingConfig()
    logging_config.update(
        log_level=log_level,
        log_output='stderr',
    )

    logger = Logger()

    templates = await load_templates(
        path,
        tags,
        file_pattern=file_pattern,
    )

    validations: list[Validator] = [ 
        bind(
            rule,
            validation,
        )
        for rule in rules.data.values()
        for _, validation in inspect.getmembers(rule)
        if isinstance(validation, Validator)
    ]
    
    if validation_error := run_validations(
        templates,
        validations,
    ):
        raise validation_error
    
    checks_passed = len(validations)
    templates_evaluated = len(templates)
    
    await logger.log(InfoLog(message=f'✅ {checks_passed} validations met for {templates_evaluated} templates'))
    


from cfn_check.rules.rules import Rules
from cfn_check.validation.validator import Validator


def bind(
    rule_set: Rules,
    validation: Validator
):
    validation.func = validation.func.__get__(rule_set, rule_set.__class__)
    setattr(rule_set, validation.func.__name__, validation.func)

    return validation




import asyncio
import os
import yaml
from cfn_check.loader.loader import (
    Loader,
    create_tag,
    find_templates,
)
from cfn_check.shared.types import YamlObject, Data

def open_template(path: str) -> YamlObject | Exception:
    try:
        with open(path, 'r') as f:
            return yaml.load(f, Loader=Loader)
    except (Exception, ) as e:
        raise e
    
def is_file(path: str) -> bool:
    return os.path.isdir(path) is False


async def load_templates(
    path: str,
    tags: list[str],
    file_pattern: str | None = None,
):
    loop = asyncio.get_event_loop()

    if await loop.run_in_executor(
        None,
        is_file,
        path,
    ) or file_pattern is None:
        template_filepaths = [
            path,
        ]

    elif file_pattern:

        template_filepaths = await loop.run_in_executor(
            None,
            find_templates,
            path,
            file_pattern,
        )

    assert len(template_filepaths) > 0 , '❌ No matching files found'
    
    for tag in tags:
        new_tag = await loop.run_in_executor(
            None,
            create_tag,
            tag,
        )

        Loader.add_constructor(f'!{tag}', new_tag)

    
    templates: list[Data]  = await asyncio.gather(*[
        loop.run_in_executor(
            None,
            open_template,
            template_path,
        ) for template_path in template_filepaths
    ])

    return templates



import textwrap
from pydantic import ValidationError
from cfn_check.rules.rule import Validator


def assemble_validation_error(
    errors: list[
        tuple[
            Validator,
            Exception | ValidationError,
        ],
    ],
) -> Exception:
    validation_error: Exception | None = None
    if len(errors) > 0:

        error_message = textwrap.indent(
            '\n'.join([
                (
                    f'Rule: {validator.name} failed\n'
                    f'Query: {validator.query}\n'
                    f'{str(err)}\n'
                )
                for validator, err in errors
            ]),
            '\t',
        )

        validation_error = Exception(
            f'\n{error_message}',
        )

    return validation_error